#[cfg(feature = "test_harness")]
use crate::{
    constants::{
        DEFAULT_EMA_PERIODS, DEFAULT_NB_CONFIRMATIONS, DEFAULT_PORT_RANGE_START,
        DEFAULT_PRICE_CHECKPOINT_INTERVAL, DEFAULT_TEST_CLOCK_TICK_MS,
        DEFAULT_TRADE_MINING_DURATION_IN_SETTLEMENT_EPOCHS, LOCALHOST_ADDRESS,
    },
    util::{
        default_trade_mining_maker_reward_percentage, default_trade_mining_reward_per_epoch,
        default_trade_mining_taker_reward_percentage,
    },
};
#[cfg(feature = "database")]
use postgres_types::{FromSql, IsNull, ToSql, Type, to_sql_checked};
#[cfg(feature = "python")]
use pyo3::prelude::*;
#[cfg(feature = "arbitrary")]
use quickcheck::Arbitrary;
use regex::Regex;
use rust_decimal::Decimal;
use serde::{Deserialize, Serialize};
use std::{
    collections::{BTreeMap, HashMap},
    convert::TryFrom,
    fmt::{self, Debug, Formatter},
    net::SocketAddr,
    str::FromStr,
    time::Duration,
};

#[cfg(all(feature = "test_harness", feature = "fixed_expiry_future"))]
use crate::constants::DEFAULT_EXPIRY_PRICE_LEAVES_DURATION;
#[cfg(feature = "test_harness")]
use crate::constants::{
    DEFAULT_FUNDING_PERIOD_LEN, DEFAULT_PNL_REALIZATION_PERIOD_LEN, DEFAULT_TRADE_MINING_PERIOD_LEN,
};
use crate::{
    Error, Result, bail,
    constants::{DEFAULT_EPOCH_LEN, DEFAULT_SETTLEMENT_EPOCH_LEN, NB_INIT_CMDS},
    ensure, error,
    ethereum_types::{Address, H256, U256},
    global::Chain,
    specs::types::{SpecsExpr, SpecsKey},
    types::{
        accounting::{Balance, Position, Price, Strategy},
        identifiers::{ReleaseHash, StrategyIdHash},
        primitives::{
            Bytes21, Bytes32, EpochId, NodeUrl, OrderSide, ProductSymbol, StampedTimeValue,
            TimeValue, TraderAddress, UnscaledI128,
        },
        request::{
            AdvanceEpoch, AdvanceSettlementEpoch, Block, Cmd, IndexPrice, MintPriceCheckpoint,
            SettlementAction, UpdateProductListings,
        },
        transaction::Ordinal,
    },
    util::serde::as_specs,
};
use ddx_common_macros::AbiToken;
use derive_builder::Builder;

#[cfg(all(feature = "fixed_expiry_future", not(target_family = "wasm")))]
use crate::util::env::EXPIRY_PRICE_LEAVES_DURATION;
#[cfg(not(target_family = "wasm"))]
use crate::{
    crypto::Keccak256,
    impl_contiguous_marker_for, impl_unsafe_byte_slice_for,
    util::{
        env::{
            EMA_PERIODS, EPOCH_SIZE, FUNDING_SETTLEMENT_MULTIPLIER, NB_CONFIRMATION,
            PNL_REALIZATION_SETTLEMENT_MULTIPLIER, PRICE_CHECKPOINT_SIZE,
            SETTLEMENT_EPOCH_MULTIPLIER, TRADE_MINING_LENGTH, TRADE_MINING_MAKER_REWARD_PERCENTAGE,
            TRADE_MINING_REWARD_PER_EPOCH, TRADE_MINING_SETTLEMENT_MULTIPLIER,
            TRADE_MINING_TAKER_REWARD_PERCENTAGE,
        },
        tokenize::Tokenizable,
    },
};
#[cfg(not(target_family = "wasm"))]
use clap::{ArgMatches, Args, Command, FromArgMatches, arg, value_parser};

mod trader;
pub use sparse_merkle_tree::H256 as StoreHash;
pub use trader::Trader;

mod tradable_product;
pub use tradable_product::{TradableProduct, TradableProductKey, TradableProductParameters};

#[cfg(not(target_family = "wasm"))]
pub type LeafMapHex = HashMap<super::primitives::Hash, (super::primitives::Hash, String)>;

/// An SMT item that can be voided, i.e. removed when determined void
pub trait VoidableItem {
    /// Check if the item is void
    fn is_void(&self) -> bool;
}

// TODO: Does this represent order_book or order_intent ?
/// A Bid or Ask in the orderbook (maker order)
#[cfg_attr(feature = "python", pyclass(get_all, set_all))]
#[derive(Debug, Clone, PartialEq, Deserialize, Serialize, AbiToken, Eq, Default)]
#[serde(rename_all = "camelCase")]
#[repr(C)]
pub struct BookOrder {
    /// Bid = 0, Ask = 1
    pub side: OrderSide,
    /// Order amount
    pub amount: UnscaledI128,
    /// Price offered to takers
    pub price: UnscaledI128,
    /// Maker address
    pub trader_address: TraderAddress,
    /// Maker strategy identifier
    pub strategy_id_hash: StrategyIdHash,
    /// Per market ordinal sequencing inclusion in the order book
    pub book_ordinal: Ordinal,
    /// Time stamp of the post order transaction
    pub time_value: TimeValue,
}

impl VoidableItem for BookOrder {
    fn is_void(&self) -> bool {
        self.amount.is_zero()
    }
}

#[cfg(feature = "python")]
#[pymethods]
impl BookOrder {
    #[new]
    fn new_py(
        side: OrderSide,
        amount: UnscaledI128,
        price: UnscaledI128,
        trader_address: TraderAddress,
        strategy_id_hash: StrategyIdHash,
        book_ordinal: Ordinal,
        time_value: TimeValue,
    ) -> Self {
        Self {
            side,
            amount,
            price,
            trader_address,
            strategy_id_hash,
            book_ordinal,
            time_value,
        }
    }
}

impl BookOrder {
    #[cfg(not(target_family = "wasm"))]
    pub fn from_order_intent(
        order_intent: &super::request::OrderIntent,
        book_ordinal: u64,
        time_value: TimeValue,
    ) -> Result<Self> {
        Ok(BookOrder {
            side: order_intent.side,
            amount: order_intent.amount,
            price: order_intent.price,
            trader_address: order_intent.trader_address()?,
            strategy_id_hash: order_intent.strategy.as_abbrev_hash(),
            book_ordinal,
            time_value,
        })
    }
}

#[cfg(feature = "arbitrary")]
impl Arbitrary for BookOrder {
    fn arbitrary(g: &mut quickcheck::Gen) -> Self {
        BookOrder {
            side: OrderSide::arbitrary(g),
            amount: UnscaledI128::arbitrary(g),
            price: UnscaledI128::arbitrary(g),
            trader_address: TraderAddress::arbitrary(g),
            strategy_id_hash: StrategyIdHash::arbitrary(g),
            book_ordinal: u64::arbitrary(g),
            time_value: u64::arbitrary(g),
        }
    }
}

/// A trader statistics value. This is used to store verified data like trade
/// volume data for trade mining.
#[cfg_attr(feature = "python", pyclass(get_all, set_all))]
#[derive(Debug, Default, Clone, PartialEq, Deserialize, Serialize, AbiToken, Eq)]
#[serde(rename_all = "camelCase")]
#[repr(C)]
pub struct Stats {
    /// The maker volume of the trader during this trade mining period
    pub maker_volume: UnscaledI128,
    /// The taker volume of the trader during this trade mining period
    pub taker_volume: UnscaledI128,
}

impl VoidableItem for Stats {
    fn is_void(&self) -> bool {
        self.maker_volume.is_zero() && self.taker_volume.is_zero()
    }
}

#[cfg(feature = "python")]
#[pymethods]
impl Stats {
    #[new]
    fn new_py(maker_volume: UnscaledI128, taker_volume: UnscaledI128) -> Self {
        Self {
            maker_volume,
            taker_volume,
        }
    }
}

#[cfg(feature = "arbitrary")]
impl Arbitrary for Stats {
    fn arbitrary(g: &mut quickcheck::Gen) -> Self {
        Self {
            maker_volume: UnscaledI128::arbitrary(g),
            taker_volume: UnscaledI128::arbitrary(g),
        }
    }
}

#[derive(Debug, Clone, PartialEq, Deserialize, Serialize, Copy)]
pub enum EpochKind {
    Regular,
    Settlement,
}

/// This epoch marker is used to track the current epochs in the state.
///
/// It can only be initialized at zero or created from an epoch command (not given arbitrary values).
#[derive(Debug, Clone, PartialEq, Deserialize, Serialize, Copy)]
#[serde(rename_all = "camelCase")]
pub struct Epoch {
    pub kind: EpochKind,
    pub epoch_id: EpochId,
    pub time: StampedTimeValue,
    pub length: u64,
}

#[cfg(feature = "arbitrary")]
impl Arbitrary for Epoch {
    fn arbitrary(g: &mut quickcheck::Gen) -> Self {
        Self {
            kind: EpochKind::Regular,
            epoch_id: u64::arbitrary(g),
            time: StampedTimeValue::arbitrary(g),
            length: u64::arbitrary(g),
        }
    }
}

impl Epoch {
    pub(super) fn new(
        kind: EpochKind,
        epoch_id: EpochId,
        time: StampedTimeValue,
        length: u64,
    ) -> Self {
        debug_assert!(time.value > 0 || (epoch_id <= 1 && time.value == 0));
        Epoch {
            kind,
            epoch_id,
            time,
            length,
        }
    }

    pub fn initial_regular(length: Option<u64>) -> Self {
        Epoch::new(
            EpochKind::Regular,
            0,
            Default::default(),
            length.unwrap_or(DEFAULT_EPOCH_LEN),
        )
    }

    pub fn initial_settlement(length: Option<u64>) -> Self {
        Epoch::new(
            EpochKind::Settlement,
            0,
            Default::default(),
            length.unwrap_or(DEFAULT_SETTLEMENT_EPOCH_LEN),
        )
    }
}

/// A leaf value in the Sparse Merkle Tree
///
/// We use a global tree to store the state. This means that all of our entity types become
/// leaves. We wrap them in this enum to mark their type.
///
/// ### Notes
///
/// The position that each entry has in this list defines the discriminant
/// that will be used when it is encoded and hashed via the AbiToken. Take
/// care to ensure that this order is identical to the order laid out in
/// the `ITEM_*` constants.
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize, AbiToken, Eq)]
pub enum Item {
    Empty,
    Trader(Trader),
    Strategy(Strategy),
    Position(Position),
    BookOrder(BookOrder),
    Price(Price),
    /// Pool of collateral populated from profitable liquidation spread.
    InsuranceFund(Balance),
    Stats(Stats),
    Signer(ReleaseHash),
    Specs(SpecsExpr),
    InsuranceFundContribution(InsuranceFundContribution),
    EpochMetadata(EpochMetadata),
    TradableProduct(TradableProduct),
}

#[cfg(feature = "python")]
pub mod exported;

#[cfg(not(target_family = "wasm"))]
impl_contiguous_marker_for!(
    Trader
    Strategy
    Position
    BookOrder
    Price
    Balance
    Stats
    ReleaseHash
    InsuranceFundContribution
    TradableProduct
);

#[cfg(not(target_family = "wasm"))]
impl_unsafe_byte_slice_for!(
    Trader
    Strategy
    Position
    BookOrder
    Price
    Balance
    Stats
    ReleaseHash
    InsuranceFundContribution
    TradableProduct
);

#[cfg(all(feature = "arbitrary", feature = "test_harness"))]
impl Arbitrary for Item {
    fn arbitrary(g: &mut quickcheck::Gen) -> Self {
        let discriminator = *g.choose(&[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]).unwrap();
        match discriminator {
            0 => Self::Empty,
            1 => Self::Trader(Arbitrary::arbitrary(g)),
            2 => Self::Strategy(Arbitrary::arbitrary(g)),
            3 => Self::Position(Arbitrary::arbitrary(g)),
            4 => Self::BookOrder(Arbitrary::arbitrary(g)),
            5 => Self::Price(Arbitrary::arbitrary(g)),
            6 => Self::InsuranceFund(Arbitrary::arbitrary(g)),
            7 => Self::Stats(Arbitrary::arbitrary(g)),
            8 => Self::Signer(Arbitrary::arbitrary(g)),
            9 => Self::Specs(Arbitrary::arbitrary(g)),
            10 => Self::InsuranceFundContribution(Arbitrary::arbitrary(g)),
            11 => Self::EpochMetadata(Arbitrary::arbitrary(g)),
            12 => Self::TradableProduct(Arbitrary::arbitrary(g)),
            _ => panic!("Invalid discriminator"),
        }
    }
}

pub const ITEM_EMPTY: u8 = 0;
pub const ITEM_TRADER: u8 = 1;
pub const ITEM_STRATEGY: u8 = 2;
pub const ITEM_POSITION: u8 = 3;
pub const ITEM_BOOK_ORDER: u8 = 4;
pub const ITEM_PRICE: u8 = 5;
pub const ITEM_INSURANCE_FUND: u8 = 6;
pub const ITEM_STATS: u8 = 7;
pub const ITEM_SIGNER: u8 = 8;
pub const ITEM_SPECS: u8 = 9;
pub const ITEM_INSURANCE_FUND_CONTRIBUTION: u8 = 10;
pub const ITEM_EPOCH_METADATA: u8 = 11;
pub const ITEM_TRADABLE_PRODUCT: u8 = 12;

impl Item {
    pub fn discriminant(&self) -> u8 {
        match self {
            Item::Empty => ITEM_EMPTY,
            Item::Trader(_) => ITEM_TRADER,
            Item::Strategy(_) => ITEM_STRATEGY,
            Item::Position(_) => ITEM_POSITION,
            Item::BookOrder(_) => ITEM_BOOK_ORDER,
            Item::Price(_) => ITEM_PRICE,
            Item::InsuranceFund(_) => ITEM_INSURANCE_FUND,
            Item::Stats(_) => ITEM_STATS,
            Item::Signer { .. } => ITEM_SIGNER,
            Item::Specs(_) => ITEM_SPECS,
            Item::InsuranceFundContribution(_) => ITEM_INSURANCE_FUND_CONTRIBUTION,
            Item::EpochMetadata(_) => ITEM_EPOCH_METADATA,
            Item::TradableProduct(_) => ITEM_TRADABLE_PRODUCT,
        }
    }
}

impl Default for Item {
    fn default() -> Self {
        Self::Empty
    }
}

#[cfg(not(target_family = "wasm"))]
impl sparse_merkle_tree::traits::Value for Item {
    fn to_h256(&self) -> sparse_merkle_tree::H256 {
        let message: Vec<u8> = if self.discriminant() == ITEM_EMPTY {
            return sparse_merkle_tree::H256::zero();
        } else {
            crate::ethabi::encode(&[self.clone().into_token()])
        };
        message.keccak256().into()
    }

    fn zero() -> Self {
        Default::default()
    }
}

#[repr(i16)]
#[derive(Debug, Clone, PartialEq, Eq, std::hash::Hash, Serialize, Deserialize)]
pub enum SealedDataKey {
    SecretUserData = 0,
}

impl TryFrom<i16> for SealedDataKey {
    type Error = ddx_common::Error;
    fn try_from(value: i16) -> core::result::Result<Self, Self::Error> {
        match value {
            0 => Ok(SealedDataKey::SecretUserData),
            _ => Err(error!("Invalid sealed data key discriminant {:?}", value)),
        }
    }
}

#[derive(Debug, Clone, Default, PartialEq, Serialize, Deserialize)]
pub struct SealedData(pub HashMap<SealedDataKey, Vec<u8>>);

impl SealedData {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn user_data(&self) -> Option<&Vec<u8>> {
        self.0.get(&SealedDataKey::SecretUserData)
    }

    pub fn insert_user_data(&mut self, data: Vec<u8>) {
        self.0.insert(SealedDataKey::SecretUserData, data);
    }
}

#[cfg(not(target_family = "wasm"))]
pub trait ConfigurableFromEnv {
    /// Read config from the environment
    fn from_env() -> Self;
}

/// Inserts initial data in the verified state tree and db then returns the state root hash.
#[derive(Debug, Clone, Deserialize, Serialize)]
pub enum BootstrapState {
    /// Runs the genesis ceremony with the unverified values.
    /// We expect these values to be published ahead of time so anyone can verify the
    /// genesis state before joining the cluster.
    #[serde(rename_all = "camelCase")]
    Genesis {
        #[serde(with = "as_specs")]
        specs: HashMap<SpecsKey, SpecsExpr>,
        #[serde(default)]
        insurance_fund_cap: Balance,
        #[serde(default)]
        ddx_fee_pool: UnscaledI128,
        #[serde(default)]
        release_hash: Bytes32,
    },
    /// Starts a cluster from an existing state.
    /// This expects the forked distributed state to be available in the local cluster.
    Resume {
        #[serde(rename = "snapshotEpochId")]
        snapshot_epoch_id: Option<EpochId>,
        #[serde(rename = "includeReleaseSigners")]
        include_release_signers: Option<ReleaseHash>,
    },
}

impl BootstrapState {
    pub fn is_genesis(&self) -> bool {
        matches!(self, BootstrapState::Genesis { .. })
    }

    #[cfg(not(target_family = "wasm"))]
    pub fn from_env_genesis() -> Self {
        let text = std::env::var("GENESIS_PARAMS").expect("GENESIS_PARAMS must be set");
        let bootstrap_state =
            serde_json::from_str::<Self>(&text).expect("Cannot deserialize GENESIS_PARAMS");
        assert!(
            bootstrap_state.is_genesis(),
            "Expected GENESIS_PARAMS to be the genesis variant"
        );
        bootstrap_state
    }
}

/// Configurable backoff parameters
#[derive(Debug, Clone)]
pub struct BackoffConfig {
    pub initial_interval: Duration,
    pub max_interval: Duration,
    pub max_elapsed_time: Duration,
}

#[cfg(feature = "test_harness")]
impl BackoffConfig {
    pub fn test_defaults() -> Self {
        BackoffConfig {
            initial_interval: Duration::from_millis(50),
            max_interval: Duration::from_secs(1),
            max_elapsed_time: Duration::from_secs(60 * 2),
        }
    }
}

#[cfg(not(target_family = "wasm"))]
impl ConfigurableFromEnv for BackoffConfig {
    fn from_env() -> Self {
        BackoffConfig {
            initial_interval: Duration::from_millis(
                std::env::var("BACKOFF_INITIAL_INTERVAL_IN_MS")
                    .ok()
                    .map(|s| s.parse::<u64>().unwrap())
                    .unwrap_or(50),
            ),
            max_interval: Duration::from_millis(
                std::env::var("BACKOFF_MAX_INTERVAL_IN_MS")
                    .ok()
                    .map(|s| s.parse::<u64>().unwrap())
                    .unwrap_or(1000),
            ),
            max_elapsed_time: Duration::from_millis(
                std::env::var("BACKOFF_MAX_ELAPSED_TIME_IN_MS")
                    .ok()
                    .map(|s| s.parse::<u64>().unwrap())
                    .unwrap_or(2 * 60 * 1000),
            ),
        }
    }
}

#[cfg(not(target_family = "wasm"))]
impl FromArgMatches for BackoffConfig {
    fn from_arg_matches(matches: &ArgMatches) -> std::prelude::v1::Result<Self, clap::Error> {
        // First build from env variables
        let mut backoff = BackoffConfig::from_env();
        // Override with values provided by arguments if applicable
        if let Some(initial_interval) = matches.get_one::<u64>("initial_interval") {
            backoff.initial_interval = Duration::from_millis(*initial_interval);
        }
        if let Some(max_interval) = matches.get_one::<u64>("max_interval") {
            backoff.max_interval = Duration::from_millis(*max_interval);
        }
        if let Some(max_elapsed_time) = matches.get_one::<u64>("max_elapsed_time") {
            backoff.max_elapsed_time = Duration::from_millis(*max_elapsed_time);
        }
        Ok(backoff)
    }

    fn update_from_arg_matches(
        &mut self,
        matches: &ArgMatches,
    ) -> std::prelude::v1::Result<(), clap::Error> {
        if let Some(initial_interval) = matches.get_one::<u64>("initial_interval") {
            self.initial_interval = Duration::from_millis(*initial_interval);
        }
        if let Some(max_interval) = matches.get_one::<u64>("max_interval") {
            self.max_interval = Duration::from_millis(*max_interval);
        }
        if let Some(max_elapsed_time) = matches.get_one::<u64>("max_elapsed_time") {
            self.max_elapsed_time = Duration::from_millis(*max_elapsed_time);
        }
        Ok(())
    }
}

#[cfg(not(target_family = "wasm"))]
impl Args for BackoffConfig {
    fn augment_args(cmd: Command) -> Command {
        cmd.arg(
            arg!(--initial_interval [MILLISECONDS] "Initial backoff interval")
                .value_parser(value_parser!(u64)),
        )
        .arg(
            arg!(--max_interval [MILLISECONDS] "Max backoff interval")
                .value_parser(value_parser!(u64)),
        )
        .arg(
            arg!(--max_elapsed_time [MILLISECONDS] "Max elapsed time before the backoff fails")
                .value_parser(value_parser!(u64)),
        )
    }

    fn augment_args_for_update(cmd: Command) -> Command {
        cmd.arg(
            arg!(--initial_interval [MILLISECONDS] "Initial backoff interval")
                .value_parser(value_parser!(u64)),
        )
        .arg(
            arg!(--max_interval [MILLISECONDS] "Max backoff interval")
                .value_parser(value_parser!(u64)),
        )
        .arg(
            arg!(--max_elapsed_time [MILLISECONDS] "Max elapsed time before the backoff fails")
                .value_parser(value_parser!(u64)),
        )
    }
}

/// List of bootstrap node URLs used for initial peer discovery
#[derive(Debug, Clone, Default)]
pub struct BootstrapNodeUrls(Vec<NodeUrl>);

impl BootstrapNodeUrls {
    pub fn new(urls: Vec<NodeUrl>) -> Self {
        Self(urls)
    }

    /// Copy into `NonEmptyNodeUrls` if not empty
    pub fn to_urls(&self) -> Result<NonEmptyNodeUrls> {
        NonEmptyNodeUrls::new(self.0.to_vec())
    }

    /// Return the inner slice of URLs
    pub fn as_urls(&self) -> &[NodeUrl] {
        &self.0
    }

    pub fn push(&mut self, url: NodeUrl) {
        self.0.push(url);
    }
}

/// List of at least one peer node URLs
#[derive(Debug, Clone)]
pub struct NonEmptyNodeUrls(Vec<NodeUrl>);

impl NonEmptyNodeUrls {
    pub fn new(urls: Vec<NodeUrl>) -> Result<Self, Error> {
        if urls.is_empty() {
            return Err(Error::Conversion(
                "At least one peer node URL must be provided".to_string(),
            ));
        }
        Ok(Self(urls))
    }

    pub fn urls(&self) -> &[NodeUrl] {
        &self.0
    }

    pub fn contains(&self, url: &NodeUrl) -> bool {
        self.0.contains(url)
    }
}

#[derive(Clone, Eq, PartialEq, Hash)]
pub enum BlockchainSender {
    UnlockedAccount(Address),
    SecretKey(H256),
}

impl BlockchainSender {
    pub fn address(&self) -> Bytes21 {
        self.eth_address().into()
    }

    pub fn eth_address(&self) -> Address {
        match self {
            BlockchainSender::UnlockedAccount(address) => *address,
            BlockchainSender::SecretKey(_secret_key) => crate::crypto::secret_key_address(
                &libsecp256k1::SecretKey::parse(_secret_key.as_fixed_bytes())
                    .expect("Private key to be a valid Secp256k1 private key"),
            ),
        }
    }

    pub fn abbrev_id(&self) -> u64 {
        let address = self.eth_address();
        let mut fixed = [0; 8];
        fixed.copy_from_slice(&address.0[0..8]);
        u64::from_be_bytes(fixed)
    }
}

/// For security reasons, printing only the Ethereum address even when using private keys.
impl fmt::Debug for BlockchainSender {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        let text = match self {
            BlockchainSender::UnlockedAccount(ref address) => {
                format!("Unlocked Account {}", address)
            }
            BlockchainSender::SecretKey(_) => "Hidden Private Key".to_string(),
        };
        f.debug_tuple("BlockchainSender").field(&text).finish()
    }
}

impl FromStr for BlockchainSender {
    type Err = Error;

    fn from_str(text: &str) -> Result<Self, Self::Err> {
        if text.len() == 42 {
            let address: Address = serde_json::from_str(format!(r#""{}""#, text).as_str())
                .map_err(|e| Error::Parse(format!("Address from str {:?}", e)))?;
            Ok(BlockchainSender::UnlockedAccount(address))
        } else if text.len() == 66 {
            let secret_key: H256 = serde_json::from_str(format!(r#""{}""#, text).as_str())
                .map_err(|e| Error::Parse(format!("H256 from str {:?}", e)))?;
            Ok(BlockchainSender::SecretKey(secret_key))
        } else {
            Err(Error::Parse(
                "Expected a 0x prefixed 20 bytes address or 32 bytes secret key".to_string(),
            ))
        }
    }
}

impl From<&Bytes21> for BlockchainSender {
    fn from(bytes: &Bytes21) -> Self {
        Self::UnlockedAccount(bytes.to_eth_address())
    }
}

impl From<Bytes21> for BlockchainSender {
    fn from(bytes: Bytes21) -> Self {
        (&bytes).into()
    }
}

#[derive(Debug, Clone, Builder)]
pub struct EnvironmentContext {
    /// Connection string for the PostgreSQL database cluster.
    pub db_connstr: String,
    /// The number of worker threads.
    pub core_threads: usize,
    /// Initial set of peer URLs for network discovery.
    pub bootstrap_nodes: BootstrapNodeUrls,
    /// The name of the deployment we are using, ie. devnet.
    pub contract_deployment: String,
    /// URL for the Contract Server that provides blockchain deployment specifications.
    pub contract_server_url: String,
    /// System backoff parameters
    pub backoff: BackoffConfig,
}

impl EnvironmentContextBuilder {
    #[cfg(not(target_family = "wasm"))]
    pub fn from_env() -> Self {
        let mut builder = Self::default();
        if let (Ok(db_name), Ok(pg_cluster)) = (
            std::env::var("OPERATOR_DB_NAME"),
            std::env::var("PG_CLUSTER"),
        ) {
            builder.db_connstr(format!("{}/{}", pg_cluster, db_name));
        }
        builder.core_threads(
            std::env::var("CORE_THREADS")
                .map(|s| s.parse().unwrap())
                // FIXME: Why not align with Enclave.xml.template?
                .unwrap_or_else(|_| crate::constants::TCS_NUM),
        );
        if let Ok(bootstrap_nodes) = std::env::var("BOOTSTRAP_NODE_URLS") {
            builder.bootstrap_nodes(BootstrapNodeUrls::new(
                bootstrap_nodes
                    .split(',')
                    .map(|s| s.parse().unwrap())
                    .collect(),
            ));
        }

        if let Ok(contract_deployment) = std::env::var("CONTRACT_DEPLOYMENT") {
            builder.contract_deployment(contract_deployment);
        }
        if let Ok(contract_server_url) = std::env::var("CONTRACT_SERVER_URL") {
            builder.contract_server_url(contract_server_url);
        }
        builder.backoff(BackoffConfig::from_env());
        builder
    }
}

#[cfg(not(target_family = "wasm"))]
impl FromArgMatches for EnvironmentContext {
    fn from_arg_matches(matches: &ArgMatches) -> std::prelude::v1::Result<Self, clap::Error> {
        // First build from env variables
        let mut builder = EnvironmentContextBuilder::from_env();
        // Override with values provided by arguments if applicable
        if let Some(db_connstr) = matches.get_one::<String>("db_connstr") {
            builder.db_connstr(db_connstr.clone());
        }
        if let Some(core_threads) = matches.get_one::<usize>("core_threads") {
            builder.core_threads(*core_threads);
        }
        if let Some(bootstrap_nodes) = matches.get_one::<String>("bootstrap_nodes") {
            builder.bootstrap_nodes(BootstrapNodeUrls::new(
                bootstrap_nodes
                    .split(',')
                    .map(|s| s.parse().unwrap())
                    .collect(),
            ));
        }
        if let Some(contract_deployment) = matches.get_one::<String>("contract_deployment") {
            builder.contract_deployment(contract_deployment.clone());
        }
        if let Some(contract_server_url) = matches.get_one::<String>("contract_server_url") {
            builder.contract_server_url(contract_server_url.clone());
        }
        if let Ok(backoff) = BackoffConfig::from_arg_matches(matches) {
            builder.backoff(backoff.clone());
        }
        Ok(builder.build().unwrap())
    }

    fn update_from_arg_matches(
        &mut self,
        matches: &ArgMatches,
    ) -> std::prelude::v1::Result<(), clap::Error> {
        if let Some(db_connstr) = matches.get_one::<String>("db_connstr") {
            self.db_connstr = db_connstr.clone();
        }
        if let Some(core_threads) = matches.get_one::<usize>("core_threads") {
            self.core_threads = *core_threads;
        }
        if let Some(bootstrap_nodes) = matches.get_one::<String>("bootstrap_nodes") {
            self.bootstrap_nodes = BootstrapNodeUrls::new(
                bootstrap_nodes
                    .split(',')
                    .map(|s| s.parse().unwrap())
                    .collect(),
            );
        }
        if let Some(contract_deployment) = matches.get_one::<String>("contract_deployment") {
            self.contract_deployment = contract_deployment.clone();
        }
        if let Some(contract_server_url) = matches.get_one::<String>("contract_server_url") {
            self.contract_server_url = contract_server_url.clone();
        }
        if let Ok(backoff) = BackoffConfig::from_arg_matches(matches) {
            self.backoff = backoff.clone();
        }
        Ok(())
    }
}

#[cfg(not(target_family = "wasm"))]
impl Args for EnvironmentContext {
    fn augment_args(cmd: Command) -> Command {
        let mut cmd = cmd;
        cmd = cmd
            .arg(arg!(--db_connstr [STRING] "Connection string for the PostgreSQL database cluster."))
            .arg( arg!(--core_threads [NUMERIC] "The number of worker threads.")
                    .value_parser(value_parser!(usize))
            )
            .arg( arg!(--bootstrap_nodes [COMMA_SEPARATED_URLS] "Initial set of peer URLs for network discovery.")
            )
            .arg(arg!(--contract_deployment [STRING] "The name of the deployment we are using, ie. devnet."))
            .arg(arg!(--contract_server_url [URL] "URL for the Contract Server that provides blockchain deployment specifications."));
        BackoffConfig::augment_args(cmd)
    }

    fn augment_args_for_update(cmd: Command) -> Command {
        let mut cmd = cmd;
        cmd = cmd
            .arg(arg!(--db_connstr [STRING] "Connection string for the PostgreSQL database cluster."))
            .arg( arg!(--core_threads [NUMERIC] "The number of worker threads.")
                    .value_parser(value_parser!(usize))
            )
            .arg( arg!(--bootstrap_nodes [COMMA_SEPARATED_URLS] "Initial set of peer URLs for network discovery.")
            )
            .arg(arg!(--contract_deployment [STRING] "The name of the deployment we are using, ie. devnet."))
            .arg(arg!(--contract_server_url [URL] "URL for the Contract Server that provides blockchain deployment specifications."));
        BackoffConfig::augment_args(cmd)
    }
}

impl From<&NodeContext> for EnvironmentContext {
    fn from(node_ctx: &NodeContext) -> Self {
        EnvironmentContext {
            db_connstr: node_ctx.db_connstr.clone(),
            core_threads: node_ctx.blocking_threads,
            bootstrap_nodes: node_ctx.bootstrap_nodes.clone(),
            contract_deployment: node_ctx.contract_deployment.clone(),
            contract_server_url: node_ctx.contract_server_url.clone(),
            backoff: node_ctx.backoff.clone(),
        }
    }
}

/// Configuration for the node.
///
/// Utilizes the builder pattern to accommodate different operating modes.
/// Primarily, `from_env` reads configurations from the environment variables,
/// while `test_defaults` provides defaults suitable for testing, eschewing assumptions
/// about default values. The builder's validation ensures all necessary parameters are provided.
#[derive(Debug, Clone, Builder)]
pub struct NodeContext {
    pub node_id: u64,
    pub custodian: BlockchainSender,
    pub db_connstr: String,
    pub bind_address: String,
    pub bind_port: u16,
    pub node_url: NodeUrl,
    pub blocking_threads: usize,
    pub nb_confirmations: u64,
    pub eth_polling_interval_in_secs: u64,
    pub contract_deployment: String,
    pub contract_server_url: String,
    pub bootstrap_nodes: BootstrapNodeUrls,
    pub want_role: WantRole,
    pub bootstrap_state: Option<BootstrapState>,
    pub backoff: BackoffConfig,
    // TODO: remove this since its only relevant for the eth-bridge
    pub submit_withdrawal_threshold: u64,
    // TODO: remove this since its only relevant for the eth-bridge
    pub submit_interval: u64,
    // TODO: Watch out with this as it skips signature validation. Place behind a feature?
    pub dev_mode: bool,
}

impl NodeContextBuilder {
    #[cfg(not(target_family = "wasm"))]
    pub fn from_env(db_env: &str) -> Self {
        use crate::constants::SUBMIT_WITHDRAWAL_THRESHOLD;

        let mut builder = Self::default();
        builder.node_url(
            std::env::var("OPERATOR_REST_API_URL")
                .unwrap_or_else(|_| "http://localhost:8080".to_string())
                .parse()
                .unwrap(),
        );
        let bind_addr =
            std::env::var("OPERATOR_BIND_ADDRESS").unwrap_or_else(|_| "localhost:8080".to_string());
        let (bind_address, bind_port) = bind_addr
            .rfind(':')
            .map(|split_at| bind_addr.split_at(split_at))
            .and_then(|(addr, port)| {
                match port.trim_matches(|c: char| !c.is_numeric()).parse::<u16>() {
                    Ok(port) => Some((addr.to_string(), port)),
                    Err(e) => {
                        tracing::error!("Couldn't parse '{}': {}", port, e);
                        None
                    }
                }
            })
            .unwrap();
        builder.bind_address(bind_address);
        builder.bind_port(bind_port);
        builder.submit_withdrawal_threshold(
            std::env::var("SUBMIT_WITHDRAWAL_THRESHOLD")
                .map(|s| s.parse().unwrap())
                .unwrap_or_else(|_| SUBMIT_WITHDRAWAL_THRESHOLD),
        );
        let daily_epochs: u64 = (Decimal::new(24 * 3600, 0) / Into::<Decimal>::into(*EPOCH_SIZE))
            .ceil()
            .try_into()
            .unwrap();
        builder.submit_interval(
            std::env::var("SUBMIT_INTERVAL")
                .map(|s| s.parse().unwrap())
                .unwrap_or_else(|_| daily_epochs),
        );
        if let Ok(node_id) = std::env::var("NODE_ID") {
            builder.node_id(node_id.parse().unwrap());
        }
        if let Ok(custodian) = std::env::var("ETH_SENDER") {
            builder.custodian(custodian.parse().unwrap());
        }
        if let (Ok(db_name), Ok(pg_cluster)) = (std::env::var(db_env), std::env::var("PG_CLUSTER"))
        {
            builder.db_connstr(format!("{}/{}", pg_cluster, db_name));
        }
        builder.blocking_threads(
            std::env::var("CORE_THREADS")
                .map(|s| s.parse().unwrap())
                // FIXME: Why not align with Enclave.xml.template?
                .unwrap_or_else(|_| crate::constants::TCS_NUM),
        );
        builder.nb_confirmations(*NB_CONFIRMATION as u64);
        builder.eth_polling_interval_in_secs(
            std::env::var("BLOCK_QUERY_POLLING_INTERVAL_IN_SECS")
                .ok()
                .map(|s| s.parse().unwrap())
                .unwrap_or(5),
        );
        if let Ok(contract_deployment) = std::env::var("CONTRACT_DEPLOYMENT") {
            builder.contract_deployment(contract_deployment);
        }
        if let Ok(contract_server_url) = std::env::var("CONTRACT_SERVER_URL") {
            builder.contract_server_url(contract_server_url);
        }
        if let Ok(bootstrap_nodes) = std::env::var("BOOTSTRAP_NODE_URLS") {
            builder.bootstrap_nodes(BootstrapNodeUrls::new(
                bootstrap_nodes
                    .split(',')
                    .map(|s| s.parse().unwrap())
                    .collect(),
            ));
        }
        // TODO: Make configurable, currently this is the only option.
        builder.want_role(WantRole::RegularNode);
        // NOTE: Defaulting to `Some(None)` (meaning a normal follower node) if BOOTSTRAP_STATE is
        // not set or set to `false`.
        builder.bootstrap_state(match std::env::var("BOOTSTRAP_STATE") {
            Ok(s) if s == "false" => None,
            Ok(s) => Some(serde_json::from_str(&s).expect("Malformed BOOTSTRAP_STATE")),
            Err(_) => None,
        });
        builder.backoff(BackoffConfig::from_env());
        // Sets false by default as it is the secure option, and only acceptable option for production nodes.
        builder.dev_mode(false);
        builder
    }

    #[cfg(feature = "test_harness")]
    pub fn test_defaults() -> Self {
        use crate::constants::TCS_NUM;

        NodeContextBuilder {
            node_id: None,
            custodian: None,
            db_connstr: None,
            bind_address: Some(LOCALHOST_ADDRESS.to_string()),
            bind_port: Some(DEFAULT_PORT_RANGE_START),
            node_url: None,
            blocking_threads: Some(TCS_NUM), // NOTE: Be careful not to starve Tokio with too low a default.
            nb_confirmations: Some(DEFAULT_NB_CONFIRMATIONS),
            eth_polling_interval_in_secs: Some(1),
            contract_deployment: None,
            contract_server_url: Some("http://snapshot:4040".to_string()),
            bootstrap_nodes: None,
            want_role: Some(WantRole::RegularNode),
            bootstrap_state: None,
            backoff: Some(BackoffConfig::test_defaults()),
            submit_withdrawal_threshold: Some(0),
            submit_interval: Some(0),
            dev_mode: Some(true),
        }
    }
}

impl NodeContext {
    pub fn bind_address(&self) -> SocketAddr {
        (format!("{}:{}", self.bind_address, self.bind_port).as_str())
            .parse()
            .unwrap()
    }
}

// Configuration variables for the trusted KYC context
#[derive(Debug, Clone, PartialEq, Default, Serialize, Deserialize)]
pub enum EthNetwork {
    #[default]
    Geth,
    Mainnet,
    Goerli,
    Sepolia,
    Gnosis,
}

impl TryInto<Chain> for EthNetwork {
    type Error = Error;

    fn try_into(self) -> Result<Chain, Self::Error> {
        match self {
            EthNetwork::Geth => Ok(Chain::Ethereum(1337)),
            EthNetwork::Mainnet => Ok(Chain::Ethereum(1)),
            EthNetwork::Goerli => Ok(Chain::Ethereum(5)),
            EthNetwork::Sepolia => Ok(Chain::Ethereum(11155111)),
            EthNetwork::Gnosis => Ok(Chain::Ethereum(100)),
        }
    }
}

impl TryFrom<Chain> for EthNetwork {
    type Error = Error;

    fn try_from(value: Chain) -> Result<Self, Self::Error> {
        match value {
            Chain::Ethereum(1337) => Ok(EthNetwork::Geth),
            Chain::Ethereum(1) => Ok(EthNetwork::Mainnet),
            Chain::Ethereum(5) => Ok(EthNetwork::Goerli),
            Chain::Ethereum(11155111) => Ok(EthNetwork::Sepolia),
            Chain::Ethereum(100) => Ok(EthNetwork::Gnosis),
            _ => Err(Error::Conversion(format!("Invalid Chain: {:?}", value))),
        }
    }
}

impl fmt::Display for EthNetwork {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            EthNetwork::Geth => write!(f, "local"),
            EthNetwork::Mainnet => write!(f, "mainnet"),
            EthNetwork::Goerli => write!(f, "goerli"),
            EthNetwork::Sepolia => write!(f, "sepolia"),
            EthNetwork::Gnosis => write!(f, "gnosis"),
        }
    }
}

impl TryFrom<&str> for EthNetwork {
    type Error = Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value {
            "mainnet" => Ok(EthNetwork::Mainnet),
            "gnosis" => Ok(EthNetwork::Gnosis),
            "goerli" => Ok(EthNetwork::Goerli),
            "sepolia" => Ok(EthNetwork::Sepolia),
            "local" => Ok(EthNetwork::Geth),
            _ => Err(Error::Conversion(format!("Invalid EthNetwork: {}", value))),
        }
    }
}
// Configuration variables for the trusted KYC context
#[derive(Debug, Default, Clone, Serialize, Deserialize)]
pub struct KycContext {
    // Eth network name
    pub network: EthNetwork,
    pub alchemy_api_key: String,
    pub blockpass_api_key: String,
}

pub fn extract_network_and_key(url: &str) -> Result<(EthNetwork, String)> {
    let pattern = r"https://eth-(?P<network>[\w-]+)\.g\.alchemy\.com/v2/(?P<key>[\w-]+)";
    let re = Regex::new(pattern).unwrap();
    if let Some(captures) = re.captures(url) {
        let network_value = captures.name("network").unwrap().as_str();
        let key_value = captures.name("key").unwrap().as_str();
        let network = EthNetwork::try_from(network_value)?;
        Ok((network, key_value.to_string()))
    } else {
        Err(Error::Conversion(
            "Failed to extract network and key from url".to_string(),
        ))
    }
}

#[cfg(not(target_family = "wasm"))]
impl ConfigurableFromEnv for KycContext {
    fn from_env() -> Self {
        let blockpass_api_key =
            std::env::var("BLOCKPASS_API_KEY").unwrap_or_else(|_| "".to_string());
        let meta = super::ethereum::DeploymentMeta::from_env();
        let app_network: EthNetwork = Chain::Ethereum(meta.chain_id).try_into().unwrap();
        if app_network != EthNetwork::Geth && app_network != EthNetwork::Gnosis {
            let (network, alchemy_api_key) = extract_network_and_key(&meta.eth_rpc_url).unwrap();
            assert!(
                network == app_network,
                "Mismatch between the network in the ETH_RPC_URL and the CHAIN_ID"
            );
            KycContext {
                network,
                alchemy_api_key,
                blockpass_api_key,
            }
        } else {
            // also use local network for Gnosis chain and Geth chain
            KycContext {
                network: app_network,
                alchemy_api_key: "".to_string(),
                blockpass_api_key,
            }
        }
    }
}

/// Contextual/configuration variables applicable to the trusted context
#[derive(Debug, Default, Clone, Serialize, Deserialize)]
pub struct TrustedContext {
    pub nb_confirmations: u32,
    pub ema_periods: u64,
    pub clock_tick_len: u64,
    // TODO: We should consider getting this from market specs.
    pub price_checkpoint_len: u64,
    pub epoch_len: u64,
    pub settlement_epoch_len: u64,
    pub static_settlement_action_periods: HashMap<SettlementAction, u64>,
    #[cfg(feature = "fixed_expiry_future")]
    pub expiry_price_leaves_duration: u64,
    pub trade_mining_len: u64,
    pub trade_mining_reward_per_epoch: UnscaledI128,
    pub trade_mining_maker_reward_percentage: UnscaledI128,
    pub trade_mining_taker_reward_percentage: UnscaledI128,
}

#[cfg(feature = "test_harness")]
impl TrustedContext {
    pub fn test_defaults() -> Self {
        TrustedContext {
            nb_confirmations: DEFAULT_NB_CONFIRMATIONS as u32,
            clock_tick_len: DEFAULT_TEST_CLOCK_TICK_MS,
            ema_periods: DEFAULT_EMA_PERIODS,
            epoch_len: DEFAULT_EPOCH_LEN,
            price_checkpoint_len: DEFAULT_PRICE_CHECKPOINT_INTERVAL,
            settlement_epoch_len: DEFAULT_SETTLEMENT_EPOCH_LEN,
            static_settlement_action_periods: [
                (
                    SettlementAction::PnlRealization,
                    DEFAULT_PNL_REALIZATION_PERIOD_LEN,
                ),
                (
                    SettlementAction::FundingDistribution,
                    DEFAULT_FUNDING_PERIOD_LEN,
                ),
                (
                    SettlementAction::TradeMining,
                    DEFAULT_TRADE_MINING_PERIOD_LEN,
                ),
            ]
            .into(),
            trade_mining_len: DEFAULT_TRADE_MINING_DURATION_IN_SETTLEMENT_EPOCHS,
            #[cfg(feature = "fixed_expiry_future")]
            expiry_price_leaves_duration: DEFAULT_EXPIRY_PRICE_LEAVES_DURATION,
            trade_mining_reward_per_epoch: default_trade_mining_reward_per_epoch().into(),
            trade_mining_maker_reward_percentage: default_trade_mining_maker_reward_percentage()
                .into(),
            trade_mining_taker_reward_percentage: default_trade_mining_taker_reward_percentage()
                .into(),
        }
    }
}

#[cfg(not(target_family = "wasm"))]
impl ConfigurableFromEnv for TrustedContext {
    fn from_env() -> Self {
        let clock_tick_len = std::env::var("CLOCK_TICK_SIZE")
            .ok()
            .map(|s| s.parse::<u64>().unwrap())
            .unwrap_or(crate::constants::DEFAULT_CLOCK_TICK_MS);

        let settlement_epoch_len = *EPOCH_SIZE * *SETTLEMENT_EPOCH_MULTIPLIER;
        TrustedContext {
            nb_confirmations: *NB_CONFIRMATION,
            ema_periods: *EMA_PERIODS,
            clock_tick_len,
            epoch_len: *EPOCH_SIZE,
            price_checkpoint_len: *PRICE_CHECKPOINT_SIZE,
            settlement_epoch_len,
            static_settlement_action_periods: [
                (
                    SettlementAction::PnlRealization,
                    settlement_epoch_len * *PNL_REALIZATION_SETTLEMENT_MULTIPLIER,
                ),
                (
                    SettlementAction::FundingDistribution,
                    settlement_epoch_len * *FUNDING_SETTLEMENT_MULTIPLIER,
                ),
                (
                    SettlementAction::TradeMining,
                    settlement_epoch_len * *TRADE_MINING_SETTLEMENT_MULTIPLIER,
                ),
            ]
            .into(),
            trade_mining_len: *TRADE_MINING_LENGTH,
            #[cfg(feature = "fixed_expiry_future")]
            expiry_price_leaves_duration: *EXPIRY_PRICE_LEAVES_DURATION,
            trade_mining_reward_per_epoch: *TRADE_MINING_REWARD_PER_EPOCH,
            trade_mining_maker_reward_percentage: *TRADE_MINING_MAKER_REWARD_PERCENTAGE,
            trade_mining_taker_reward_percentage: *TRADE_MINING_TAKER_REWARD_PERCENTAGE,
        }
    }
}

#[cfg(feature = "database")]
impl ToSql for Epoch {
    fn to_sql(
        &self,
        ty: &Type,
        out: &mut bytes::BytesMut,
    ) -> Result<IsNull, Box<dyn std::error::Error + 'static + Sync + Send>> {
        let json: String = serde_json::to_string(&self)?;
        json.to_sql(ty, out)
    }

    fn accepts(ty: &Type) -> bool {
        <String as ToSql>::accepts(ty)
    }

    to_sql_checked!();
}

#[cfg(feature = "database")]
impl<'a> FromSql<'a> for Epoch {
    fn from_sql(
        ty: &Type,
        raw: &'a [u8],
    ) -> Result<Self, Box<dyn std::error::Error + Sync + Send>> {
        let s: String = String::from_sql(ty, raw)?;
        let decoded: Epoch = serde_json::from_str(&s)?;
        Ok(decoded)
    }

    fn accepts(ty: &Type) -> bool {
        <String as FromSql>::accepts(ty)
    }
}

#[derive(AbiToken, Debug)]
pub struct LatestCheckpoint {
    pub block_number: U256,
    pub state_root: H256,
    pub transaction_root: H256,
    pub epoch_id: U256,
}

/// Request log commands for the transition of each epoch type.
///
/// This is generally discovered by find the last command of each type in the request log.
/// Both the sequencer and processor need this to know when to advance to the next epoch.
#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
pub struct TickCmds {
    pub epoch_cmd: AdvanceEpoch,
    pub settlement_cmd: AdvanceSettlementEpoch,
    pub price_checkpoint_cmd: MintPriceCheckpoint,
    pub update_product_listings_cmd: UpdateProductListings,
}

#[derive(Debug, Default, Clone, Serialize, Deserialize, PartialEq)]
pub struct AllCmds {
    pub time: StampedTimeValue,
    pub block_cmd: Block,
    pub epoch_cmd: AdvanceEpoch,
    pub settlement_cmd: AdvanceSettlementEpoch,
    pub price_checkpoint_cmd: MintPriceCheckpoint,
    pub update_product_listings_cmd: UpdateProductListings,
    pub index_price_cmds: Vec<IndexPrice>,
}

impl AllCmds {
    pub fn split(self) -> (TickCmds, TimeValue, Block, Vec<IndexPrice>) {
        let AllCmds {
            time,
            epoch_cmd,
            settlement_cmd,
            price_checkpoint_cmd,
            update_product_listings_cmd,
            block_cmd,
            index_price_cmds,
        } = self;
        (
            TickCmds {
                epoch_cmd,
                settlement_cmd,
                price_checkpoint_cmd,
                update_product_listings_cmd,
            },
            time.value,
            block_cmd,
            index_price_cmds,
        )
    }
}

impl TryFrom<Vec<Cmd>> for AllCmds {
    type Error = Error;

    fn try_from(value: Vec<Cmd>) -> Result<Self, Self::Error> {
        // TODO: Do we need to check number of IndexPrice? If so then need env var specifying the
        // number of products expected
        ensure!(
            value.iter().fold(0, |acc, x| {
                if !matches!(x, Cmd::IndexPrice(_) | Cmd::Genesis,) {
                    acc + 1
                } else {
                    acc
                }
            }) == NB_INIT_CMDS,
            "Expected {:?} non-index-price commands, found {:?}",
            NB_INIT_CMDS,
            value
        );
        let mut all_cmds = AllCmds::default();
        let mut index_price_cmds = vec![];
        for cmd in value {
            match cmd {
                Cmd::Block(b) => all_cmds.block_cmd = b,
                Cmd::AdvanceTime(t) => all_cmds.time = t,
                Cmd::AdvanceEpoch(e) => all_cmds.epoch_cmd = e,
                Cmd::AdvanceSettlementEpoch(s) => all_cmds.settlement_cmd = s,
                Cmd::PriceCheckpoint(c) => all_cmds.price_checkpoint_cmd = c,
                Cmd::IndexPrice(p) => index_price_cmds.push(p),
                Cmd::UpdateProductListings(u) => all_cmds.update_product_listings_cmd = u,
                _ => bail!("Unexpected command {:?}", cmd),
            }
        }
        all_cmds.index_price_cmds = index_price_cmds;
        Ok(all_cmds)
    }
}

// TODO: Link with the codes
#[derive(Debug, Clone, PartialEq, Eq, std::hash::Hash, Serialize, Deserialize)]
pub enum CmdKind {
    AdvanceEpoch,
    AdvanceSettlementEpoch,
    MintPriceCheckpoint,
    AdvanceTime,
    AdvanceBlock,
    UpdateProductListings,
}

/// Holds commands at the tail of the request log.
///
/// The tail represents the latest (or current) commands for each kind. For example, to query the current
/// epoch at the request log level, we fetch the `TailCmds` and select the `CmdKind::AdvanceEpoch` entry.
#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
pub struct TailCmds {
    pub cmds: AllCmds,
    pub next_index_to_assign: u64,
}

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
pub struct EpochInitParams {
    // TODO: Consider storing the sequenced request list and cast in the enclave while verifying signatures
    pub cmds: AllCmds,
    /// Index of the last request to process, may exist in the request log but not the tx log.
    pub next_index_to_process: u64,
}

pub type SidedBook = BTreeMap<Decimal, Vec<BookOrder>>;

#[derive(Debug, Clone, PartialEq, Deserialize, Serialize, AbiToken)]
pub enum WantRole {
    /// The node is completely passive; replicating entries, but neither voting nor timing out.
    Learner,
    /// Regular node replicating logs from the leader and may be elected as leader.
    RegularNode,
}

/// An insurance fund contribution
#[cfg_attr(feature = "python", pyclass(get_all, set_all))]
#[derive(Debug, Clone, Default, PartialEq, Serialize, Deserialize, AbiToken, Eq)]
#[serde(rename_all = "camelCase")]
#[repr(C)]
pub struct InsuranceFundContribution {
    /// Total contribution to the insurance fund, minus any withdrawals
    pub avail_balance: Balance,
    /// Balance locked for withdrawal
    pub locked_balance: Balance,
}

impl VoidableItem for InsuranceFundContribution {
    fn is_void(&self) -> bool {
        self.avail_balance.is_zero() && self.locked_balance.is_zero()
    }
}

#[cfg(feature = "python")]
#[pymethods]
impl InsuranceFundContribution {
    #[new]
    fn new_py(avail_balance: Balance, locked_balance: Balance) -> Self {
        Self {
            avail_balance,
            locked_balance,
        }
    }
}

#[cfg(feature = "arbitrary")]
impl Arbitrary for InsuranceFundContribution {
    fn arbitrary(g: &mut quickcheck::Gen) -> Self {
        Self {
            avail_balance: Arbitrary::arbitrary(g),
            locked_balance: Arbitrary::arbitrary(g),
        }
    }
}

/// Metadata about an epoch that has been transitioned.
#[cfg_attr(feature = "python", pyclass(get_all, set_all))]
#[derive(Debug, Clone, Default, PartialEq, Serialize, Deserialize, AbiToken, Eq)]
#[serde(rename_all = "camelCase")]
pub struct EpochMetadata {
    /// The DDX fee pool. This represents the total DDX fees collected in this epoch and are
    /// distributed among custodians during a valid checkpoint.
    pub ddx_fee_pool: UnscaledI128,
    /// The next book ordinals for all of the active markets. This mapping
    /// allows follower nodes to reliably generate the market index from the
    /// sparse merkle tree. The ordinals are logged at the end of the epoch and are None if this is
    /// the current epoch.
    pub next_book_ordinals: HashMap<ProductSymbol, u64>,
}

impl VoidableItem for EpochMetadata {
    fn is_void(&self) -> bool {
        false
    }
}

#[cfg(feature = "python")]
#[pymethods]
impl EpochMetadata {
    #[new]
    fn new_py(ddx_fee_pool: UnscaledI128, next_book_ordinals: HashMap<ProductSymbol, u64>) -> Self {
        Self {
            ddx_fee_pool,
            next_book_ordinals,
        }
    }
}

#[cfg(feature = "arbitrary")]
impl Arbitrary for EpochMetadata {
    fn arbitrary(g: &mut quickcheck::Gen) -> Self {
        Self {
            ddx_fee_pool: Arbitrary::arbitrary(g),
            next_book_ordinals: Arbitrary::arbitrary(g),
        }
    }
}

#[derive(PartialEq, PartialOrd)]
pub struct SchemaVersion {
    pub major: u64,
    pub minor: u64,
    pub patch: u64,
}

impl FromStr for SchemaVersion {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut version_array = [0_u64; 3];
        if !s.is_empty() {
            for (i, v) in s.splitn(3, '.').enumerate() {
                version_array[i] = v
                    .parse::<u64>()
                    .map_err(|_| Error::Parse(format!("Failed to parse version schema {}", s)))?;
            }
        }
        Ok(SchemaVersion {
            major: version_array[0],
            minor: version_array[1],
            patch: version_array[2],
        })
    }
}

impl std::fmt::Debug for SchemaVersion {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}.{}.{}", self.major, self.minor, self.patch)
    }
}

impl std::fmt::Display for SchemaVersion {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

#[cfg(feature = "database")]
impl ToSql for SchemaVersion {
    fn to_sql(
        &self,
        ty: &Type,
        out: &mut bytes::BytesMut,
    ) -> Result<IsNull, Box<dyn std::error::Error + 'static + Sync + Send>> {
        self.to_string().to_sql(ty, out)
    }

    fn accepts(ty: &Type) -> bool {
        <String as ToSql>::accepts(ty)
    }

    to_sql_checked!();
}

#[cfg(feature = "database")]
impl<'a> FromSql<'a> for SchemaVersion {
    fn from_sql(
        ty: &Type,
        raw: &'a [u8],
    ) -> Result<Self, Box<dyn std::error::Error + Sync + Send>> {
        let version_string: String = String::from_sql(ty, raw)?;
        Ok(version_string.parse::<Self>()?)
    }

    fn accepts(ty: &Type) -> bool {
        <String as FromSql>::accepts(ty)
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{ethabi::Token, util::tokenize::generate_schema};
    use quickcheck::QuickCheck;

    #[test]
    fn prop_ethabi_roundtrip_item() {
        fn ethabi_roundtrip_item(input: Item) -> bool {
            println!("input: {:?}", input);
            let token = input.clone().into_token();
            let bytes = crate::ethabi::encode(&[token.clone()]);
            let schema = generate_schema(&token);
            let decoded_token: Token = crate::ethabi::decode(&[schema.into()], &bytes)
                .unwrap()
                .remove(0);
            let decoded = Item::from_token(decoded_token).unwrap();
            decoded == input
        }
        QuickCheck::new()
            .gen(quickcheck::Gen::new(10))
            .quickcheck(ethabi_roundtrip_item as fn(Item) -> bool);
    }

    #[test]
    fn parse_eth_rpc_url() {
        let url = "https://eth-mainnet.g.alchemy.com/v2/1234567890abcdef";
        let (network, key) = extract_network_and_key(url).unwrap();
        assert_eq!(network, EthNetwork::Mainnet);
        assert_eq!(key, "1234567890abcdef");
    }

    #[test]
    fn parse_schema_version() {
        let empty_version = "";
        let schema_version = empty_version.parse::<SchemaVersion>().unwrap();
        assert_eq!(schema_version.to_string(), String::from("0.0.0"));

        let version_string = "1.2.3";
        let schema_version = version_string.parse::<SchemaVersion>().unwrap();
        assert_eq!(
            schema_version,
            SchemaVersion {
                major: 1,
                minor: 2,
                patch: 3
            }
        );
    }
}
